#
# Copyright (c) 2018 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import re
import codecs
from enum import IntEnum, unique

import fnutil


WIDTH_MAX = 127
HEIGHT_MAX = 255
SWIDTH_MAX = 32000

class Width:
	def __init__(self, x, y):
		self.x = x
		self.y = y


	@staticmethod
	def _parse(name, value, limit_x, limit_y):
		words = fnutil.split_words(name, value, 2)
		return Width(fnutil.parse_dec('width x', words[0], -limit_x, limit_x),
			fnutil.parse_dec('width y', words[1], -limit_y, limit_y))


	@staticmethod
	def parse_s(value):
		return Width._parse('SWIDTH', value, SWIDTH_MAX, SWIDTH_MAX)


	@staticmethod
	def parse_d(value):
		return Width._parse('DWIDTH', value, WIDTH_MAX, HEIGHT_MAX)


OFFSET_MIN = -128
OFFSET_MAX = 127

class BBX:
	def __init__(self, width, height, xoff, yoff):
		self.width = width
		self.height = height
		self.xoff = xoff
		self.yoff = yoff


	@staticmethod
	def parse(name, value):
		words = fnutil.split_words(name, value, 4)
		return BBX(fnutil.parse_dec('width', words[0], 1, WIDTH_MAX),
			fnutil.parse_dec('height', words[1], 1, HEIGHT_MAX),
			fnutil.parse_dec('bbxoff', words[2], -WIDTH_MAX, WIDTH_MAX),
			fnutil.parse_dec('bbyoff', words[3], -WIDTH_MAX, WIDTH_MAX))


	def row_size(self):
		return (self.width + 7) >> 3


	def __str__(self):
		return '%d %d %d %d' % (self.width, self.height, self.xoff, self.yoff)


class Props:
	def __init__(self):
		self.names = []
		self.values = []


	def add(self, name, value):
		self.names.append(name)
		self.values.append(value)


	def clone(self):
		props = Props()
		props.names = self.names[:]
		props.values = self.values[:]
		return props


	def get(self, name):
		try:
			return self.values[self.names.index(name)]
		except ValueError:
			return None


	class Iter:
		def __init__(self, props):
			self.index = 0
			self.props = props


		def __next__(self):
			if self.index == len(self.props.names):
				raise StopIteration

			result = (self.props.names[self.index], self.props.values[self.index])
			self.index += 1
			return result


	def __iter__(self):
		return Props.Iter(self)


	def parse(self, line, name, callback=None):
		if not line or not line.startswith(bytes(name, 'ascii')):
			raise Exception(name + ' expected')

		value = line[len(name):].lstrip()
		self.add(name, value)
		return value if callback is None else callback(name, value)


	def set(self, name, value):
		try:
			self.values[self.names.index(name)] = value
		except ValueError:
			self.add(name, value)


class Base:
	def __init__(self):
		self.props = Props()
		self.bbx = None
		self.finis = []


	def keep_comments(self, line):
		if not line.startswith(b'COMMENT'):
			return line

		self.props.add('', line)
		return None


	def keep_finishes(self, line):
		self.finis.append(line)
		return None if line.startswith(b'COMMENT') else line


class Char(Base):
	def __init__(self):
		Base.__init__(self)
		self.code = -1
		self.swidth = None
		self.dwidth = None
		self.data = None


	@staticmethod
	def bitmap(data, row_size):
		bitmap = ''

		for index in range(0, len(data), row_size):
			bitmap += data[index : index + row_size].hex() + '\n'

		return bytes(bitmap, 'ascii').upper()


	def _read(self, input):
		# HEADER
		read_next = lambda: input.read_lines(lambda line: self.keep_comments(line))
		read_prop = lambda name, callback=None: self.props.parse(read_next(), name, callback)

		read_prop('STARTCHAR')
		self.code = read_prop('ENCODING', fnutil.parse_dec)
		self.swidth = read_prop('SWIDTH', lambda _, value: Width.parse_s(value))
		self.dwidth = read_prop('DWIDTH', lambda _, value: Width.parse_d(value))
		self.bbx = read_prop('BBX', BBX.parse)
		line = read_next()

		if line and line.startswith(b'ATTRIBUTES'):
			self.props.parse(line, 'ATTRIBUTES')
			line = read_next()

		# BITMAP
		if self.props.parse(line, 'BITMAP') != b'':
			raise Exception('BITMAP expected')

		row_len = self.bbx.row_size() * 2
		bitmap = b''

		for _ in range(0, self.bbx.height):
			line = read_next()

			if not line:
				raise Exception('bitmap data expected')

			if len(line) == row_len:
				bitmap += line
			else:
				raise Exception('invalid bitmap length')

		# FINAL
		if input.read_lines(lambda line: self.keep_finishes(line)) != b'ENDCHAR':
			raise Exception('ENDCHAR expected')

		self.data = codecs.decode(bitmap, 'hex')  # no spaces allowed
		return self


	@staticmethod
	def read(input):
		return Char()._read(input)  # pylint: disable=protected-access


	def write(self, output):
		for [name, value] in self.props:
			output.write_prop(name, value)

		output.write_line(Char.bitmap(self.data, self.bbx.row_size()) + b'\n'.join(self.finis))


@unique
class XLFD(IntEnum):
	FOUNDRY = 1
	FAMILY_NAME = 2
	WEIGHT_NAME = 3
	SLANT = 4
	SETWIDTH_NAME = 5
	ADD_STYLE_NAME = 6
	PIXEL_SIZE = 7
	POINT_SIZE = 8
	RESOLUTION_X = 9
	RESOLUTION_Y = 10
	SPACING = 11
	AVERAGE_WIDTH = 12
	CHARSET_REGISTRY = 13
	CHARSET_ENCODING = 14

CHARS_MAX = 65535

class Font(Base):
	def __init__(self):
		Base.__init__(self)
		self.xlfd = []
		self.chars = []
		self.default_code = -1


	def get_ascent(self):
		ascent = self.props.get('FONT_ASCENT')

		if ascent is not None:
			return fnutil.parse_dec('FONT_ASCENT', ascent, -HEIGHT_MAX, HEIGHT_MAX)

		return self.bbx.height + self.bbx.yoff


	def get_bold(self):
		return int(b'bold' in self.xlfd[XLFD.WEIGHT_NAME].lower())


	def get_italic(self):
		return int(re.search(b'^[IO]', self.xlfd[XLFD.SLANT]) is not None)


	def _read(self, input):
		# HEADER
		read_next = lambda: input.read_lines(lambda line: self.keep_comments(line))
		read_prop = lambda name, callback=None: self.props.parse(read_next(), name, callback)
		line = input.read_lines(Font.skip_empty)

		if self.props.parse(line, 'STARTFONT') != b'2.1':
			raise Exception('STARTFONT 2.1 expected')

		self.xlfd = read_prop('FONT', lambda name, value: value.split(b'-', 15))

		if len(self.xlfd) != 15 or self.xlfd[0] != b'':
			raise Exception('non-XLFD font names are not supported')

		read_prop('SIZE')
		self.bbx = read_prop('FONTBOUNDINGBOX', BBX.parse)
		line = read_next()

		if line and line.startswith(b'STARTPROPERTIES'):
			num_props = self.props.parse(line, 'STARTPROPERTIES', fnutil.parse_dec)

			for _ in range(0, num_props):
				line = read_next()

				if not line:
					raise Exception('property expected')

				match = re.fullmatch(br'(\w+)\s+([-\d"].*)', line)

				if not match:
					raise Exception('invalid property format')

				name = str(match.group(1), 'ascii')
				value = match.group(2)

				if name == 'DEFAULT_CHAR':
					self.default_code = fnutil.parse_dec(name, value)

				self.props.add(name, value)

			if read_prop('ENDPROPERTIES') != b'':
				raise Exception('ENDPROPERTIES expected')

			line = read_next()

		# GLYPHS
		num_chars = self.props.parse(line, 'CHARS', lambda name, value: fnutil.parse_dec(name, value, 1, CHARS_MAX))

		for _ in range(0, num_chars):
			self.chars.append(Char.read(input))

		if next((char.code for char in self.chars if char.code == self.default_code), -1) != self.default_code:
			raise Exception('invalid DEFAULT_CHAR')

		# FINAL
		if input.read_lines(lambda line: self.keep_finishes(line)) != b'ENDFONT':
			raise Exception('ENDFONT expected')

		if input.read_lines(Font.skip_empty):
			raise Exception('garbage after ENDFONT')

		return self


	@staticmethod
	def read(input):
		return Font()._read(input)  # pylint: disable=protected-access


	@staticmethod
	def skip_empty(line):
		return line if line else None


	def write(self, output):
		for [name, value] in self.props:
			output.write_prop(name, value)

		for char in self.chars:
			char.write(output)

		output.write_line(b'\n'.join(self.finis))



#
# Copyright (c) 2019 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import re
import copy

import fnutil
import fncli
import fnio
import bdf


class Params(fncli.Params):
	def __init__(self):
		fncli.Params.__init__(self)
		self.filter = False
		self.family = None
		self.output = None


HELP = ('' +
	'usage: ucstoany [-f] [-F FAMILY] [-o OUTPUT] INPUT REGISTRY ENCODING TABLE...\n' +
	'Generate a BDF font subset.\n' +
	'\n' +
	'  -f, --filter   Discard characters with unicode FFFF; with registry ISO10646,\n' +
	'                 encode the first 32 characters with their indexes; with other\n' +
	'                 registries, encode all characters with indexes\n' +
	'  -F FAMILY      output font family name (default = input)\n' +
	'  -o OUTPUT      output file (default = stdout)\n' +
	'  TABLE          text file, one hexadecimal unicode per line\n' +
	'  --help         display this help and exit\n' +
	'  --version      display the program version and license, and exit\n' +
	'  --excstk       display the exception stack on error\n' +
	'\n' +
	'The input must be a BDF 2.1 font encoded in the unicode range.\n' +
	'Any COMMENT-s are discarded, duplicate properties are collapsed.\n' +
	'Unlike ucs2any, all TABLE-s form a single subset of the input font.\n')

VERSION = 'ucstoany 1.55, Copyright (C) 2019 Dimitar Toshkov Zhekov\n\n' + fnutil.GPL2PLUS_LICENSE

class Options(fncli.Options):
	def __init__(self):
		fncli.Options.__init__(self, ['-F', '-o'], HELP, VERSION)


	def parse(self, name, value, params):
		if name in ['-f', '--filter']:
			params.filter = True
		elif name == '-F':
			params.family = bytes(value, 'ascii')
			if '-' in value:
				raise Exception('family name may not contain "-"')
		elif name == '-o':
			params.output = value
		else:
			self.fallback(name, params)


def main_program(nonopt, parsed):
	bstr = lambda number: bytes(str(number), 'ascii')

	# NON-OPTIONS
	if len(nonopt) < 4:
		raise Exception('invalid number of arguments, try --help')

	input = nonopt[0]
	registry = nonopt[1]
	encoding = nonopt[2]
	new_codes = []

	if not re.fullmatch(r'[A-Za-z][\w.:()]*', registry) or not re.fullmatch(r'[\w.:()]+', encoding):
		raise Exception('invalid registry or encoding')

	# READ INPUT
	ifs = fnio.InputStream(input)

	try:
		old_font = bdf.Font.read(ifs)
		ifs.close()
	except Exception as ex:
		raise Exception(ifs.location() + str(ex))

	# READ TABLES
	def load_code(line):
		new_codes.append(fnutil.parse_hex('unicode', line))

	for name in nonopt[3:]:
		ifs = fnio.InputStream(name)

		try:
			ifs.read_lines(load_code)
			ifs.close()
		except Exception as ex:
			raise Exception(ifs.location() + str(ex))

	if not new_codes:
		raise Exception('no characters in the output font')

	# CREATE GLYPHS
	new_font = bdf.Font()
	charmap = {char.code:char for char in old_font.chars}
	index = 0
	unstart = 0
	family = parsed.family if parsed.family is not None else old_font.xlfd[bdf.XLFD.FAMILY_NAME]

	if parsed.filter:
		unstart = 32 if registry == 'ISO10646' else bdf.CHARS_MAX

	for code in new_codes:
		if code == 0xFFFF and parsed.filter:
			index += 1
			continue

		if code in charmap:
			old_char = charmap[code]
			uni_ffff = False
		else:
			uni_ffff = True

			if code != 0xFFFF:
				raise Exception('%s does not contain U+%04X' % (input, code))

			if old_font.default_code != -1:
				old_char = charmap[old_font.default_code]
			elif 0xFFFD in charmap:
				old_char = charmap[0xFFFD]
			else:
				raise Exception('%s does not contain U+FFFF, and no replacement found' % input)

		new_char = copy.copy(old_char)
		new_char.code = code if index >= unstart else index
		index += 1
		new_char.props = new_char.props.clone()
		new_char.props.set('ENCODING', bstr(new_char.code))
		new_font.chars.append(new_char)

		if uni_ffff:
			new_char.props.set('STARTCHAR', b'uniFFFF')
		elif old_char.code == old_font.default_code or (old_char.code == 0xFFFD and new_font.default_code == -1):
			new_font.default_code = new_char.code

	# CREATE HEADER
	registry = bytes(registry, 'ascii')
	encoding = bytes(encoding, 'ascii')

	for [name, value] in old_font.props:
		if name == 'FONT':
			new_font.xlfd = old_font.xlfd[:]
			new_font.xlfd[bdf.XLFD.FAMILY_NAME] = family
			new_font.xlfd[bdf.XLFD.CHARSET_REGISTRY] = registry
			new_font.xlfd[bdf.XLFD.CHARSET_ENCODING] = encoding
			value = b'-'.join(new_font.xlfd)
		elif name == 'STARTPROPERTIES':
			num_props = fnutil.parse_dec(name, value, 1)
		elif name == 'FAMILY_NAME':
			value = fnutil.quote(family)
		elif name == 'CHARSET_REGISTRY':
			value = fnutil.quote(registry)
		elif name == 'CHARSET_ENCODING':
			value = fnutil.quote(encoding)
		elif name == 'DEFAULT_CHAR':
			if new_font.default_code != -1:
				value = bstr(new_font.default_code)
			else:
				num_props -= 1
				continue
		elif name == 'ENDPROPERTIES':
			if new_font.default_code != -1 and new_font.props.get('DEFAULT_CHAR') is None:
				new_font.props.add('DEFAULT_CHAR', bstr(new_font.default_code))
				num_props += 1

			new_font.props.set('STARTPROPERTIES', bstr(num_props))
		elif name == 'CHARS':
			value = bstr(len(new_font.chars))

		new_font.props.add(name, value)

	# COPY FIELDS
	new_font.bbx = old_font.bbx
	new_font.finis = old_font.finis

	# WRITE OUTPUT
	ofs = fnio.OutputStream(parsed.output)

	try:
		new_font.write(ofs)
		ofs.close()
	except Exception as ex:
		raise Exception(ofs.location() + str(ex) + ofs.destroy())


if __name__ == '__main__':
	fncli.start('ucstoany.py', Options(), Params(), main_program)



#
# Copyright (c) 2018 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import bdf


class Char:
	def __init__(self, code, name, width, data):
		self.code = code
		self.name = name
		self.width = width
		self.data = data


	@staticmethod
	def from_bdf(char, fbbox):
		delta_yoff = char.bbx.yoff - fbbox.yoff  # ~DSB

		if delta_yoff < 0:
			raise Exception('char %d: BBX yoff < FONTBOUNDINGBOX yoff' % char.code)

		if char.dwidth.x >= 0:
			if char.bbx.xoff >= 0:
				width = max(char.bbx.width + char.bbx.xoff, char.dwidth.x)
				dst_xoff = char.bbx.xoff
			else:
				width = max(char.bbx.width, char.dwidth.x - char.bbx.xoff)
				dst_xoff = 0
		else:
			dst_xoff = max(char.bbx.xoff - char.dwidth.x, 0)
			width = char.bbx.width + dst_xoff

		if width > bdf.WIDTH_MAX:
			raise Exception('char %d: output width > %d' % (char.code, bdf.WIDTH_MAX))

		height = fbbox.height
		src_row_size = char.bbx.row_size()
		dst_row_size = (width + 7) >> 3
		dst_ymax = height - delta_yoff
		dst_ymin = dst_ymax - char.bbx.height
		compat_row = dst_xoff == 0 and width >= char.bbx.width

		if compat_row and src_row_size == dst_row_size and dst_ymin == 0 and dst_ymax == height:
			data = char.data
		elif dst_ymin < 0:
			raise Exception('char %d: start row %d' % (char.code, dst_ymin))
		elif compat_row:
			src_byte_no = 0
			data = bytearray(dst_ymin * dst_row_size)
			line_fill = bytes(dst_row_size - src_row_size)

			for dst_y in range(dst_ymin, dst_ymax):
				data += char.data[src_byte_no : src_byte_no + src_row_size] + line_fill
				src_byte_no += src_row_size

			data += bytes(delta_yoff * dst_row_size)
		else:
			data = bytearray(dst_row_size * height)

			for dst_y in range(dst_ymin, dst_ymax):
				src_byte_no = (dst_y - dst_ymin) * src_row_size
				dst_byte_no = dst_y * dst_row_size + (dst_xoff >> 3)

				src_bit_no = 7
				dst_bit_no = 7 - (dst_xoff & 7)

				for _ in range(0, char.bbx.width):
					if char.data[src_byte_no] & (1 << src_bit_no):
						data[dst_byte_no] |= (1 << dst_bit_no)

					if src_bit_no > 0:
						src_bit_no -= 1
					else:
						src_bit_no = 7
						src_byte_no += 1

					if dst_bit_no > 0:
						dst_bit_no -= 1
					else:
						dst_bit_no = 7
						dst_byte_no += 1

		return Char(char.code, char.props.get('STARTCHAR'), width, data)


	def row_size(self):
		return (self.width + 7) >> 3


	def write(self, output, max_width, yoffset):
		output.write_line(b'STARTCHAR %s\nENCODING %d' % (self.name, self.code))
		output.write_line(b'SWIDTH %d 0\nDWIDTH %d 0' % (round(self.width * 1000 / max_width), self.width))
		output.write_line(b'BBX %d %d 0 %d' % (self.width, len(self.data) / self.row_size(), yoffset))
		output.write_line(b'BITMAP\n' + bdf.Char.bitmap(self.data, self.row_size()) + b'ENDCHAR')


class Font(bdf.Font):
	def __init__(self):
		bdf.Font.__init__(self)
		self.min_width = bdf.WIDTH_MAX
		self.avg_width = 0


	def _read(self, input):
		bdf.Font._read(self, input)
		self.chars = [Char.from_bdf(char, self.bbx) for char in self.chars]
		self.bbx.xoff = 0
		total_width = 0

		for char in self.chars:
			self.min_width = min(self.min_width, char.width)
			self.bbx.width = max(self.bbx.width, char.width)
			total_width += char.width

		self.avg_width = round(total_width / len(self.chars))
		self.props.set('FONTBOUNDINGBOX', bytes(str(self.bbx), 'ascii'))
		return self


	@staticmethod
	def read(input):
		return Font()._read(input)  # pylint: disable=protected-access


	def get_proportional(self):
		return int(self.bbx.width > self.min_width)


	def write(self, output):
		for [name, value] in self.props:
			output.write_prop(name, value)

		for char in self.chars:
			char.write(output, self.bbx.width, self.bbx.yoff)

		output.write_line(b'ENDFONT')



#
# Copyright (c) 2019 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import re

import fnutil
import fncli
import fnio
import bmpf


class Params(fncli.Params):
	def __init__(self):
		fncli.Params.__init__(self)
		self.version = -1
		self.exchange = -1
		self.output = None


HELP = ('' +
	'usage: bdftopsf [-1|-2|-r] [-g|-G] [-o OUTPUT] [INPUT.bdf] [TABLE...]\n' +
	'Convert a BDF font to PC Screen Font or raw font\n' +
	'\n' +
	'  -1, -2      write a PSF version 1 or 2 font (default = 1 if possible)\n' +
	'  -r, --raw   write a RAW font\n' +
	'  -g, --vga   exchange the characters at positions 0...31 with these at\n' +
	'              192...223 (default for VGA text mode compliant PSF fonts\n' +
	'              with 224 to 512 characters starting with unicode 00A3)\n' +
	'  -G          do not exchange characters 0...31 and 192...223\n' +
	'  -o OUTPUT   output file (default = stdout, must not be a terminal)\n' +
	'  --help      display this help and exit\n' +
	'  --version   display the program version and license, and exit\n' +
	'  --excstk    display the exception stack on error\n' +
	'\n' +
	'The input must be a monospaced unicode-encoded BDF 2.1 font.\n' +
	'\n' +
	'The tables are text files with two or more hexadecimal unicodes per line:\n' +
	'a character code from the BDF, and extra code(s) for it. All extra codes\n' +
	'are stored sequentially in the PSF unicode table for their character.\n' +
	'<ss> is always specified as FFFE, although it is stored as FE in PSF2.\n')

VERSION = 'bdftopsf 1.50, Copyright (C) 2019 Dimitar Toshkov Zhekov\n\n' + fnutil.GPL2PLUS_LICENSE

class Options(fncli.Options):
	def __init__(self):
		fncli.Options.__init__(self, ['-o'], HELP, VERSION)


	def parse(self, name, value, params):
		if name in ['-1', '-2']:
			params.version = int(name[1])
		elif name in ['-r', '--raw']:
			params.version = 0
		elif name in ['-g', '--vga']:
			params.exchange = True
		elif name == '-G':
			params.exchange = False
		elif name == '-o':
			params.output = value
		else:
			self.fallback(name, params)


def main_program(nonopt, parsed):
	version = parsed.version
	exchange = parsed.exchange
	bdfile = len(nonopt) > 0 and nonopt[0].lower().endswith('.bdf')
	ver1_unicodes = True

	# READ INPUT
	ifs = fnio.InputStream(nonopt[0] if bdfile else None)

	try:
		font = bmpf.Font.read(ifs)
		ifs.close()

		for char in font.chars:
			prefix = 'char %d: ' % char.code

			if char.width != font.bbx.width:
				raise Exception(prefix + 'output width not equal to maximum output width')

			if char.code == 65534:
				raise Exception(prefix + 'not a character, use 65535 for empty position')

			if char.code >= 65536:
				if version == 1:
					raise Exception(prefix + '-1 requires unicodes <= 65535')
				ver1_unicodes = False

		# VERSION
		ver1_num_chars = len(font.chars) == 256 or len(font.chars) == 512

		if version == 1:
			if not ver1_num_chars:
				raise Exception('-1 requires a font with 256 or 512 characters')

			if font.bbx.width != 8:
				raise Exception('-1 requires a font with width 8')

		# EXCHANGE
		vga_num_chars = len(font.chars) >= 224 and len(font.chars) <= 512
		vga_text_size = font.bbx.width == 8 and font.bbx.height in [8, 14, 16]

		if exchange is True:
			if not vga_num_chars:
				raise Exception('-g/--vga requires a font with 224...512 characters')

			if not vga_text_size:
				raise Exception('-g/--vga requires an 8x8, 8x14 or 8x16 font')

	except Exception as ex:
		raise Exception(ifs.location() + str(ex))

	# READ TABLES
	tables = dict()

	def load_extra(line):
		nonlocal ver1_unicodes

		words = re.split(br'\s+', line)

		if len(words) < 2:
			raise Exception('invalid format')

		uni = fnutil.parse_hex('unicode', words[0])

		if uni == 0xFFFE:
			raise Exception('FFFE is not a character')

		if next((char for char in font.chars if char.code == uni), None):
			if uni >= 0x10000:
				ver1_unicodes = False

			if uni not in tables:
				tables[uni] = []

			table = tables[uni]

			for word in words[1:]:
				dup = fnutil.parse_hex('extra code', word)

				if dup == 0xFFFF:
					raise Exception('FFFF is not a character')

				if dup >= 0x10000:
					ver1_unicodes = False

				if not dup in table or 0xFFFE in table:
					tables[uni].append(dup)

			if version == 1 and not ver1_unicodes:
				raise Exception('-1 requires unicodes <= FFFF')

	for name in nonopt[int(bdfile):]:
		ifs = fnio.InputStream(name)

		try:
			ifs.read_lines(load_extra)
			ifs.close()
		except Exception as ex:
			raise Exception(ifs.location() + str(ex))

	# VERSION
	if version == -1:
		version = 1 if ver1_num_chars and ver1_unicodes and font.bbx.width == 8 else 2

	# EXCHANGE
	if exchange == -1:
		exchange = vga_text_size and version >= 1 and vga_num_chars and font.chars[0].code == 0x00A3

	if exchange:
		font.chars = font.chars[192:224] + font.chars[32:192] + font.chars[0:32] + font.chars[224:]

	# WRITE
	ofs = fnio.OutputStream(parsed.output)

	if ofs.file.isatty():
		raise Exception('binary output may not be send to a terminal, use -o or redirect/pipe it')

	try:
		# HEADER
		if version == 1:
			ofs.write8(0x36)
			ofs.write8(0x04)
			ofs.write8((len(font.chars) >> 8) + 1)
			ofs.write8(font.bbx.height)
		elif version == 2:
			ofs.write32(0x864AB572)
			ofs.write32(0x00000000)
			ofs.write32(0x00000020)
			ofs.write32(0x00000001)
			ofs.write32(len(font.chars))
			ofs.write32(len(font.chars[0].data))
			ofs.write32(font.bbx.height)
			ofs.write32(font.bbx.width)

		# GLYPHS
		for char in font.chars:
			ofs.write(char.data)

		# UNICODES
		if version > 0:
			def write_unicode(code):
				if version == 1:
					ofs.write16(code)
				elif code <= 0x7F:
					ofs.write8(code)
				elif code in [0xFFFE, 0xFFFF]:
					ofs.write8(code & 0xFF)
				else:
					if code <= 0x7FF:
						ofs.write8(0xC0 + (code >> 6))
					else:
						if code <= 0xFFFF:
							ofs.write8(0xE0 + (code >> 12))
						else:
							ofs.write8(0xF0 + (code >> 18))
							ofs.write8(0x80 + ((code >> 12) & 0x3F))

						ofs.write8(0x80 + ((code >> 6) & 0x3F))

					ofs.write8(0x80 + (code & 0x3F))

			for char in font.chars:
				if char.code != 0xFFFF:
					write_unicode(char.code)

				if char.code in tables:
					for extra in tables[char.code]:
						write_unicode(extra)

				write_unicode(0xFFFF)

		# FINISH
		ofs.close()

	except Exception as ex:
		raise Exception(ofs.location() + str(ex) + ofs.destroy())


if __name__ == '__main__':
	fncli.start('bdftopsf.py', Options(), Params(), main_program)



#
# Copyright (c) 2018 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import sys
import re


UNICODE_MAX = 1114111  # 0x10FFFF


def parse_dec(name, s, min_value=0, max_value=UNICODE_MAX):
	try:
		value = int(s)
	except ValueError:
		raise Exception('invalid %s format' % name)

	if min_value is not None and value < min_value:
		raise Exception('%s must be >= %d' % (name, min_value))

	if max_value is not None and value > max_value:
		raise Exception('%s must be <= %d' % (name, max_value))

	return value


def parse_hex(name, s, min_value=0, max_value=UNICODE_MAX):
	try:
		value = int(s, 16)
	except ValueError:
		raise Exception('invalid %s format' % name)

	if min_value is not None and value < min_value:
		raise Exception('%s must be >= %X' % (name, min_value))

	if max_value is not None and value > max_value:
		raise Exception('%s must be <= %X' % (name, max_value))

	return value


def quote(bstr):
	return b'"%s"' % bstr.replace(b'"', b'""')


def unquote(bstr, name=None):
	if len(bstr) >= 2 and bstr.startswith(b'"') and bstr.endswith(b'"'):
		bstr = bstr[1 : len(bstr) - 1].replace(b'""', b'"')
	elif name is not None:
		raise Exception(name + ' must be quoted')

	return bstr


def warning(prefix, message):
	if prefix.endswith(':'):
		prefix += ' '
	elif prefix and not prefix.endswith(': '):
		prefix += ': '

	sys.stderr.write('%swarning: %s\n' % (prefix, message))


def split_words(name, bstr, count):
	words = re.split(br'\s+', bstr, count)

	if len(words) != count:
		raise Exception('%s must contain %d values' % (name, count))

	return words


GPL2PLUS_LICENSE = ('' +
	'This program is free software; you can redistribute it and/or\n' +
	'modify it under the terms of the GNU General Public License as\n' +
	'published by the Free Software Foundation; either version 2 of\n' +
	'the License, or (at your option) any later version.\n' +
	'\n' +
	'This program is distributed in the hope that it will be useful,\n' +
	'but WITHOUT ANY WARRANTY; without even the implied warranty of\n' +
	'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n' +
	'GNU General Public License for more details.\n')



#
# Copyright (c) 2019 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import re

import fnutil
import fncli
import fnio
import bdf
import bmpf


class Params(fncli.Params):
	def __init__(self):
		fncli.Params.__init__(self)
		self.char_set = -1
		self.min_char = -1
		self.fnt_family = 0
		self.output = None


HELP = ('' +
	'usage: bdftofnt [-c CHARSET] [-m MINCHAR] [-f FAMILY] [-o OUTPUT] [INPUT]\n' +
	'Convert a BDF font to Windows FNT\n' +
	'\n' +
	'  -c CHARSET   fnt character set (default = 0, see wingdi.h ..._CHARSET)\n' +
	'  -m MINCHAR   fnt minimum character code (8-bit CP decimal, not unicode)\n' +
	'  -f FAMILY    fnt family: DontCare, Roman, Swiss, Modern or Decorative\n' +
	'  -o OUTPUT    output file (default = stdout, must not be a terminal)\n' +
	'  --help       display this help and exit\n' +
	'  --version    display the program version and license, and exit\n' +
	'  --excstk     display the exception stack on error\n' +
	'\n' +
	'The input must be a BDF font encoded in the unicode range.\n')

VERSION = 'bdftofnt 1.55, Copyright (C) 2019 Dimitar Toshkov Zhekov\n\n' + fnutil.GPL2PLUS_LICENSE

FNT_FAMILIES = ['DontCare', 'Roman', 'Swiss', 'Modern', 'Decorative']

class Options(fncli.Options):
	def __init__(self):
		fncli.Options.__init__(self, ['-c', '-m', '-f', '-o'], HELP, VERSION)


	def parse(self, name, value, params):
		if name == '-c':
			params.char_set = fnutil.parse_dec('charset', value, 0, 255)
		elif name == '-m':
			params.min_char = fnutil.parse_dec('minchar', value, 0, 255)
		elif name == '-f':
			if value in FNT_FAMILIES:
				params.fnt_family = FNT_FAMILIES.index(value)
			else:
				raise Exception('invalid fnt family')
		elif name == '-o':
			params.output = value
		else:
			self.fallback(name, params)


WIN_FONTHEADERSIZE = 118
FNT_CHARSETS = [238, 204, 0, 161, 162, 177, 178, 186, 163]

def main_program(nonopt, parsed):
	if len(nonopt) > 1:
		raise Exception('invalid number of arguments, try --help')

	char_set = parsed.char_set
	min_char = parsed.min_char

	# READ INPUT
	ifs = fnio.InputStream(nonopt[0] if nonopt else None)

	try:
		font = bmpf.Font.read(ifs)
		ifs.close()
	except Exception as ex:
		raise Exception(ifs.location() + str(ex))

	# COMPUTE
	if char_set == -1:
		encoding = font.xlfd[bdf.XLFD.CHARSET_ENCODING]

		if re.fullmatch(b'(cp)?125[0-8]', encoding.lower()):
			char_set = FNT_CHARSETS[int(encoding[-1:])]
		else:
			char_set = 255

	try:
		num_chars = len(font.chars)

		if num_chars > 256:
			raise Exception('too many characters, the maximum is 256')

		if min_char == -1:
			if num_chars in [192, 256]:
				min_char = 256 - num_chars
			else:
				min_char = font.chars[0].code

		max_char = min_char + num_chars - 1

		if max_char >= 256:
			raise Exception('the maximum character code is too big, (re)specify -m')

		# HEADER
		vtell = WIN_FONTHEADERSIZE + (num_chars + 1) * 4
		bits_offset = vtell
		ctable = []
		width_bytes = 0

		# CTABLE/GLYPHS
		for char in font.chars:
			row_size = char.row_size()
			ctable.append(char.width)
			ctable.append(vtell)
			vtell += row_size * font.bbx.height
			width_bytes += row_size

		if vtell > 0xFFFF:
			raise Exception('too much character data')

		# SENTINEL
		sentinel = 2 - width_bytes % 2
		ctable.append(sentinel * 8)
		ctable.append(vtell)
		vtell += sentinel * font.bbx.height
		width_bytes += sentinel

		if width_bytes > 0xFFFF:
			raise Exception('the total character width is too big')

	except Exception as ex:
		raise Exception(ifs.location() + str(ex))

	# WRITE
	ofs = fnio.OutputStream(parsed.output)

	if ofs.file.isatty():
		raise Exception('binary output may not be send to a terminal, use -o or redirect/pipe it')

	try:
		# HEADER
		family = font.xlfd[bdf.XLFD.FAMILY_NAME]
		copyright = font.props.get('COPYRIGHT')
		copyright = fnutil.unquote(copyright)[:60] if copyright is not None else b''
		proportional = font.get_proportional()

		ofs.write16(0x0200)                                              # font version
		ofs.write32(vtell + len(family) + 1)                             # total size
		ofs.write_zstr(copyright, 60 - len(copyright))
		ofs.write16(0)                                                   # gdi, device type
		ofs.write16(round(font.bbx.height * 72 / 96))
		ofs.write16(96)                                                  # vertical resolution
		ofs.write16(96)                                                  # horizontal resolution
		ofs.write16(font.get_ascent())                                   # base line
		ofs.write16(0)                                                   # internal leading
		ofs.write16(0)                                                   # external leading
		ofs.write8(font.get_italic())
		ofs.write8(0)                                                    # underline
		ofs.write8(0)                                                    # strikeout
		ofs.write16(400 + 300 * font.get_bold())
		ofs.write8(char_set)
		ofs.write16(0 if proportional else font.bbx.width)
		ofs.write16(font.bbx.height)
		ofs.write8((parsed.fnt_family << 4) + proportional)
		ofs.write16(font.avg_width)
		ofs.write16(font.bbx.width)
		ofs.write8(min_char)
		ofs.write8(max_char)

		default_index = max_char - min_char
		break_index = 0

		if font.default_code != -1:
			default_index = next(index for index, char in enumerate(font.chars) if char.code == font.default_code)

		if min_char <= 0x20 <= max_char:
			break_index = 0x20 - min_char

		ofs.write8(default_index)
		ofs.write8(break_index)
		ofs.write16(width_bytes)
		ofs.write32(0)            # device name
		ofs.write32(vtell)
		ofs.write32(0)            # gdi bits pointer
		ofs.write32(bits_offset)
		ofs.write8(0)             # reserved

		# CTABLE
		for value in ctable:
			ofs.write16(value)

		# GLYPHS
		data = bytearray(font.bbx.height * font.bbx.row_size())

		for char in font.chars:
			row_size = char.row_size()
			counter = 0
			# MS coordinates
			for n in range(0, row_size):
				for y in range(0, font.bbx.height):
					data[counter] = char.data[row_size * y + n]
					counter += 1
			ofs.write(data[:counter])
		ofs.write(bytes(sentinel * font.bbx.height))

		# FAMILY
		ofs.write_zstr(family, 1)
		ofs.close()

	except Exception as ex:
		raise Exception(ofs.location() + str(ex) + ofs.destroy())


if __name__ == '__main__':
	fncli.start('bdftofnt.py', Options(), Params(), main_program)



#
# Copyright (c) 2018 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import codecs
import struct
import sys
import os


BINARY_ENCODING = '_fnio_binary'

class InputStream:
	def __init__(self, file_name, encoding=BINARY_ENCODING):
		if file_name is not None:
			if encoding == BINARY_ENCODING:
				self.file = open(file_name, 'rb')
			else:
				self.file = open(file_name, 'r', encoding=encoding)

			self.st_name = file_name
		else:
			if encoding == BINARY_ENCODING:
				self.file = sys.stdin.buffer
			elif encoding is None:
				self.file = sys.stdin
			else:
				self.file = codecs.getreader(encoding)(sys.stdin.buffer)

			self.st_name = '<stdin>'

		self.line_no = 0
		self.eof = False


	def close(self):
		self.line_no = 0
		self.eof = False
		self.file.close()


	def location(self):
		return '%s:%s' % (self.st_name, 'EOF: ' if self.eof else '%d: ' % self.line_no if self.line_no > 0 else ' ')


	def read_line(self):
		return self.read_lines(lambda line: line)


	def read_lines(self, callback):
		try:
			for line in self.file:
				self.line_no += 1
				self.eof = False
				line = callback(line.rstrip())
				if line is not None:
					return line
		except OSError:
			self.line_no = 0
			self.eof = False
			raise

		self.eof = True
		return None


class OutputStream:
	def __init__(self, file_name):
		if file_name is not None:
			self.file = open(file_name, 'wb')
			self.st_name = file_name
		else:
			self.file = sys.stdout.buffer
			self.st_name = '<stdout>'

		self.close_attempt = False


	def close(self):
		self.close_attempt = True
		self.file.close()


	def destroy(self):
		errors = ''

		if self.file != sys.stdout.buffer:
			if not self.close_attempt:
				try:
					self.file.close()
				except Exception as ex:
					errors += '\n%s: close: %s' % (self.st_name, str(ex))

			try:
				os.remove(self.st_name)
			except Exception as ex:
				errors += '\n%s: unlink: %s' % (self.st_name, str(ex))

		return errors


	def location(self):
		return self.st_name + ': '


	def write(self, buffer):
		self.file.write(buffer)


	def write8(self, value):
		self.file.write(struct.pack('B', value))


	def write16(self, value):
		self.file.write(struct.pack('<H', value))


	def write32(self, value):
		self.file.write(struct.pack('<L', value))


	def write_line(self, bstr):
		self.file.write(bstr + b'\n')


	def write_prop(self, name, value):
		self.write_line((bytes(name, 'ascii') + b' ' + value).strip())


	def write_zstr(self, bstr, num_zeros):
		self.file.write(bstr + bytes(num_zeros))



#
# Copyright (c) 2019 Dimitar Toshkov Zhekov <dimitar.zhekov@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

import sys
import os
import re


class Params:
	def __init__(self):
		self.excstk = False


class Options:
	def __init__(self, need_args, help_text, version_text):
		for name in need_args:
			if not re.fullmatch('(-[^-]|--[^=]+)', name):
				raise Exception('invalid option name "%s"' % name)

		self.need_args = need_args
		self.help_text = help_text
		self.version_text = version_text


	def posixly_correct(self):  # pylint: disable=no-self-use
		return 'POSIXLY_CORRECT' in os.environ


	def needs_arg(self, name):
		return name in self.need_args


	def fallback(self, name, params):
		if name == '--excstk':
			params.excstk = True
		elif name == '--help' and self.help_text is not None:
			sys.stdout.write(self.help_text)
			sys.exit(0)
		elif name == '--version' and self.version_text is not None:
			sys.stdout.write(self.version_text)
			sys.exit(0)
		else:
			suffix = ' (taking an argument?)' if self.needs_arg(name) else ''
			suffix += ', try --help' if self.help_text is not None else ''
			raise Exception('unknown option "%s"%s' % (name, suffix))


	def reader(self, args, skip_zero=True):
		return Options.Reader(self, args, skip_zero)


	class Reader:
		def __init__(self, options, args, skip_zero):
			self.options = options
			self.args = args
			self.skip_zero = skip_zero


		def __iter__(self):
			return Options.Reader.Iterator(self)


		class Iterator:
			def __init__(self, reader):
				self.options = reader.options
				self.args = reader.args
				self.optind = int(reader.skip_zero)
				self.chrind = 1
				self.endopt = False


			def __next__(self):
				if self.chrind == 0:
					self.optind += 1
					self.chrind = 1

				if self.optind == len(self.args):
					raise StopIteration

				arg = self.args[self.optind]

				if self.endopt or arg == '-' or not arg.startswith('-'):
					self.endopt = self.options.posixly_correct()
					name = None
					value = arg
				elif arg == '--':
					self.chrind = 0
					self.endopt = True
					return next(self)
				elif not arg.startswith('--'):
					name = '-' + arg[self.chrind]
					self.chrind += 1
					if self.chrind < len(arg):
						if not self.options.needs_arg(name):
							return (name, None)
						value = arg[self.chrind:]
					else:
						value = None
				elif '=' in arg and arg.index('=') >= 3:
					name = arg.split('=', 1)[0]
					if not self.options.needs_arg(name):
						raise Exception('option "%s" does not take an argument' % name)
					value = arg[len(name) + 1:]
				else:
					name = arg
					value = None

				if value is None and int(self.options.needs_arg(name)) > 0:
					self.optind += 1
					if self.optind == len(self.args):
						raise Exception('option "%s" requires an argument' % name)
					value = self.args[self.optind]

				self.chrind = 0
				return (name, value)


def start(program_name, options, params, main_program):
	parsed = Params() if params is None else params

	try:
		if sys.hexversion < 0x3050000:
			raise Exception('python 3.5.0 or later required')

		if params is None:
			return main_program(options.reader(sys.argv), lambda name: options.fallback(name, parsed))

		nonopt = []

		for [name, value] in options.reader(sys.argv):
			if name is None:
				nonopt.append(value)
			else:
				options.parse(name, value, parsed)

		return main_program(nonopt, parsed)

	except Exception as ex:
		if parsed.excstk:
			raise

		sys.stderr.write('%s: %s\n' % (sys.argv[0] if sys.argv[0] else program_name, str(ex)))
		sys.exit(1)
